"""
Aqara Protocol

Provide protocol level support for Aqara, feature including
- Join multicast group
- Receive / Send messages
- Encoding / Decoding messages
- Utility to unicast / broadcast messages
"""

import json
import logging
import socket
import struct

from aqara.const import (MCAST_ADDR, MCAST_PORT, GATEWAY_PORT)

_LOGGER = logging.getLogger(__name__)

class AqaraProtocol(object):
    """Base aqara client protocol."""

    def __init__(self):
        self.transport = None

    def connection_made(self, transport):
        """Implementation when connection is made."""
        self.transport = transport
        self._add_membership()

    def connection_lost(self, exc):
        """Implementation when connection is lost."""
        _LOGGER.info("Disconnected")
        _LOGGER.error(exc)

    def datagram_received(self, data, addr):
        """Implementation when datagram is received."""
        data_str = data.decode('utf-8')
        msg = json.loads(data_str)
        _LOGGER.debug('recv: %s', data_str)
        self.handle_message(msg, addr)

    def error_received(self, exc):
        """Implentation when error is received."""
        _LOGGER.error('error_received: %s', exc)

    def handle_message(self, msg, src_addr):
        """Callback to handle new messages, override to add implementation."""
        _LOGGER.debug('handle_message from %s: %s', src_addr, json.dumps(msg))

    def broadcast(self, msg):
        """Send a message to the Aqara multicast channel."""
        self._send(msg, (MCAST_ADDR, MCAST_PORT))

    def unicast(self, addr, msg):
        """Send a message to a specific gateway at <ip>"""
        self._send(msg, (addr, GATEWAY_PORT))

    def _send(self, msg, dest):
        """private: send a message as UDP packet."""
        _LOGGER.debug('send: %s', json.dumps(msg))
        data = json.dumps(msg).encode('utf-8')
        self.transport.sendto(data, dest)

    def _add_membership(self):
        """private: add multicast membership"""
        _LOGGER.debug("Joining multicast group...")
        sock = self.transport.get_extra_info("socket")
        group = socket.inet_aton(MCAST_ADDR)
        mreq = struct.pack("4sL", group, socket.INADDR_ANY)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        _LOGGER.debug("Multicast membership added")



"""Aqara Devices"""

import json
import logging

from pydispatch import dispatcher
from aqara.const import (
    AQARA_DEVICE_HT,
    AQARA_DEVICE_MOTION,
    AQARA_DEVICE_MAGNET,
    AQARA_DEVICE_SWITCH,
    AQARA_SWITCH_ACTION_CLICK,
    AQARA_SWITCH_ACTION_DOUBLE_CLICK,
    AQARA_SWITCH_ACTION_LONG_CLICK_PRESS,
    AQARA_SWITCH_ACTION_LONG_CLICK_RELEASE,
    AQARA_DEFAULT_VOLTAGE
)

HASS_UPDATE_SIGNAL = "update_hass_sensor"
_LOGGER = logging.getLogger(__name__)

BUTTON_ACTION_MAP = {
    "click": AQARA_SWITCH_ACTION_CLICK,
    "double_click": AQARA_SWITCH_ACTION_DOUBLE_CLICK,
    "long_click_press": AQARA_SWITCH_ACTION_LONG_CLICK_PRESS,
    "long_click_release": AQARA_SWITCH_ACTION_LONG_CLICK_RELEASE
}

def create_device(gateway, model, sid):
    """Device factory"""
    if model == AQARA_DEVICE_HT:
        return AqaraHTSensor(gateway, sid)
    elif model == AQARA_DEVICE_MOTION:
        return AqaraMotionSensor(gateway, sid)
    elif model == AQARA_DEVICE_MAGNET:
        return AqaraContactSensor(gateway, sid)
    elif model == AQARA_DEVICE_SWITCH:
        return AqaraSwitchSensor(gateway, sid)
    else:
        raise RuntimeError('Unsupported device type: {} [{}]'.format(model, sid))

class AqaraBaseDevice(object):
    """AqaraBaseDevice"""
    def __init__(self, model, gateway, sid):
        self._gateway = gateway
        self._model = model
        self._sid = sid
        self._device_props = {}

    @property
    def sid(self):
        """property: sid"""
        return self._sid

    @property
    def model(self):
        """property: model"""
        return self._model

    def subscribe_update(self, handle_update):
        """subscribe to sensor update event"""
        dispatcher.connect(handle_update, signal=HASS_UPDATE_SIGNAL, sender=self)

    def unsubscribe_update(self, handle_update):
        """unsubscribe from sensor update event"""
        dispatcher.disconnect(handle_update, signal=HASS_UPDATE_SIGNAL, sender=self)

    def update_now(self):
        """force read sensor data"""
        self._gateway.read_device(self._sid)

    def on_update(self, data):
        """handler for sensor data update"""
        self.do_update(data)
        self.log_info("update " + json.dumps(self._device_props))
        dispatcher.send(signal=HASS_UPDATE_SIGNAL, sender=self)

    def on_heartbeat(self, data):
        """handler for heartbeat"""
        self.do_heartbeat(data)
        dispatcher.send(signal=HASS_UPDATE_SIGNAL, sender=self)

    def do_update(self, data):
        """update sensor state according to data"""
        pass

    def do_heartbeat(self, data):
        """update heartbeat"""
        pass

    def log_warning(self, msg):
        """log warning"""
        self._log(_LOGGER.warning, msg)

    def log_info(self, msg):
        """log info"""
        self._log(_LOGGER.info, msg)

    def log_debug(self, msg):
        """log debug"""
        self._log(_LOGGER.debug, msg)

    def _log(self, log_func, msg):
        """log"""
        log_func('%s [%s]: %s', self.sid, self.model, msg)

class AqaraHTSensor(AqaraBaseDevice):
    """AqaraHTSensor"""
    def __init__(self, gateway, sid):
        super().__init__(AQARA_DEVICE_HT, gateway, sid)
        self._device_props = {
            "temperature": 0,
            "humidity": 0
        }

    @property
    def temperature(self):
        """property: temperature (unit: C)"""
        return self._device_props["temperature"]

    @property
    def humidity(self):
        """property: humidity (unit: %)"""
        return self._device_props["humidity"]

    def do_update(self, data):
        if "temperature" in data:
            self._device_props["temperature"] = self.parse_value(data["temperature"])
        if "humidity" in data:
            self._device_props["humidity"] = self.parse_value(data["humidity"])

    def do_heartbeat(self, data):
        # heartbeat for HT sensor contains the same data as report
        self.do_update(data)

    @staticmethod
    def parse_value(str_value):
        """parse sensor_ht values"""
        return round(int(str_value) / 100, 1)


class AqaraContactSensor(AqaraBaseDevice):
    """AqaraContactSensor"""
    def __init__(self, gateway, sid):
        super().__init__(AQARA_DEVICE_MAGNET, gateway, sid)
        self._device_props = {
            "triggered": False,
            "voltage": AQARA_DEFAULT_VOLTAGE
        }

    @property
    def triggered(self):
        """property: triggered (bool)"""
        return self._device_props["triggered"]

    def do_update(self, data):
        if "status" in data:
            self._device_props["triggered"] = data["status"] == "open"

    def do_heartbeat(self, data):
        if "voltage" in data:
            self._device_props["voltage"] = int(data["voltage"])

class AqaraMotionSensor(AqaraBaseDevice):
    """AqaraMotionSensor"""
    def __init__(self, gateway, sid):
        super().__init__(AQARA_DEVICE_MOTION, gateway, sid)
        self._device_props = {
            "triggered": False,
            "voltage": AQARA_DEFAULT_VOLTAGE
        }

    @property
    def triggered(self):
        """property: triggered (bool)"""
        return self._device_props["triggered"]

    def do_update(self, data):
        if "status" in data:
            self._device_props["triggered"] = data["status"] == "motion"
        else:
            self._device_props["triggered"] = False

    def do_heartbeat(self, data):
        if "voltage" in data:
            self._device_props["voltage"] = int(data["voltage"])

class AqaraSwitchSensor(AqaraBaseDevice):
    """AqaraMotionSensor"""
    def __init__(self, gateway, sid):
        super().__init__(AQARA_DEVICE_SWITCH, gateway, sid)
        self._device_props = {
            "action": None,
            "voltage": AQARA_DEFAULT_VOLTAGE
        }

    @property
    def action(self):
        """property: last_action"""
        return self._device_props["action"]

    def do_update(self, data):
        if "status" in data:
            status = data["status"]
            if status in BUTTON_ACTION_MAP:
                self._device_props["action"] = BUTTON_ACTION_MAP[status]
            else:
                self.log_warning('invalid status: {}' % status)

    def do_heartbeat(self, data):
        if "voltage" in data:
            self._device_props["voltage"] = int(data["voltage"])



"""
Aqara Client

A client implementation to receive Aqara events and send control messages.

Feature includes:
- Discover gateways
- Discover devices of a gateway
- Subscribe to updates for devices
- Read values from a device (async)
- Send control command to a device (async)
- Heartbeat

"""
import asyncio
import json
import logging

from pydispatch import dispatcher
from aqara.protocol import AqaraProtocol
from aqara.gateway import AqaraGateway
from aqara.const import (
    LISTEN_IP, LISTEN_PORT,
    AQARA_EVENT_NEW_GATEWAY
)

_LOGGER = logging.getLogger(__name__)

def _extract_data(msg):
    return json.loads(msg["data"])

class AqaraClient(AqaraProtocol):
    """Aqara Client implementation."""
    def __init__(self, gw_secrets=None):
        super().__init__()
        self.transport = None
        self._gw_secrets = {} if gw_secrets is None else gw_secrets
        self._gateways = {}
        self._device_to_gw = {}

    @property
    def gateways(self):
        """property: gateways"""
        return self._gateways

    @asyncio.coroutine
    def start(self, loop):
        """Start listening on gateway events"""
        listen = loop.create_datagram_endpoint(lambda: self, local_addr=(LISTEN_IP, LISTEN_PORT))
        transport, _protocol = yield from listen
        self.transport = transport
        self.discover_gateways()
        _LOGGER.info("started")

    def stop(self):
        """Stop listening to gateway events"""
        if self.transport is None:
            _LOGGER.info("not started")
        else:
            self.transport.close()
            _LOGGER.info("stopped")

    def discover_gateways(self):
        """Ask all gateways to respond identity."""
        _LOGGER.info('discovering gateways...')
        discovery_msg = {"cmd": "whois"}
        self.broadcast(discovery_msg)

    def discover_devices(self, gw_addr):
        """Ask a gateway to reply with the SID of all attached devices."""
        discover_devices_msg = {"cmd": "get_id_list"}
        self.unicast(gw_addr, discover_devices_msg)

    def read_device(self, gw_addr, sid):
        """Send a request to read device 'sid' on gateway 'gw_addr'"""
        read_msg = {"cmd": "read", "sid": sid}
        self.unicast(gw_addr, read_msg)

    def write_device(self, gw_addr, model, sid, data, meta=None):
        """Send a request to write 'data' to device 'sid' on gateway 'gw_addr'"""
        write_msg = {
            "cmd": "write",
            "model": model,
            "sid": sid,
            "data": json.dumps(data)
        }

        if meta != None:
            write_msg.update(meta)

        self.unicast(gw_addr, write_msg)

    def handle_message(self, msg, src_addr):
        """Override: handle_message implementation"""
        _LOGGER.debug("handle_message from %s", src_addr)

        cmd = msg["cmd"]
        sid = msg["sid"]

        if cmd == "iam":
            addr = msg["ip"]
            self.on_gateway_discovered(sid, addr)
        elif cmd == "get_id_list_ack":
            sids = _extract_data(msg)
            self.on_devices_discovered(sid, sids)
        elif cmd == "read_ack":
            model = msg["model"]
            data = _extract_data(msg)
            self.on_read_ack(model, sid, data)
        elif cmd == "write_ack":
            if "model" not in msg:
                _LOGGER.error("write error: %s", json.dumps(msg))
                return
            model = msg["model"]
            data = _extract_data(msg)
            self.on_write_ack(model, sid, data)
        elif cmd == "report":
            model = msg["model"]
            data = _extract_data(msg)
            self.on_report(model, sid, data)
        elif cmd == "heartbeat":
            model = msg["model"]
            data = _extract_data(msg)
            gw_token = None if "token" not in msg else msg["token"]
            self.on_heartbeat(model, sid, data, gw_token)

    def on_gateway_discovered(self, gw_sid, gw_addr):
        """Called when a gateway is discovered"""
        _LOGGER.info("discovered gateway at %s [%s]", gw_sid, gw_addr)
        gw_secret = None
        if gw_sid in self._gw_secrets:
            gw_secret = self._gw_secrets[gw_sid]
        new_gateway = AqaraGateway(self, gw_sid, gw_addr, gw_secret)
        self._gateways[gw_sid] = new_gateway
        self._device_to_gw[gw_sid] = new_gateway
        dispatcher.send(signal=AQARA_EVENT_NEW_GATEWAY, gateway=new_gateway, sender=self)

    def on_devices_discovered(self, gw_sid, sids):
        """Called when list of devices of gateway is returned."""
        if gw_sid not in self._gateways:
            _LOGGER.error("on_devices_discovered(): gateway %s not found", gw_sid)
            return
        gateway = self._gateways[gw_sid]
        for sid in sids:
            _LOGGER.info("found device %s on gateway %s", sid, gw_sid)
            self._device_to_gw[sid] = gateway
        gateway.on_devices_discovered(sids)

    def on_read_ack(self, model, sid, data):
        """Called when a gateway send back ACK for a read request."""
        if sid not in self._device_to_gw:
            _LOGGER.error("on_read_ack(): sid not found %s", sid)
            return

        self._device_to_gw[sid].on_read_ack(model, sid, data)

    def on_write_ack(self, model, sid, data):
        """Called when a gateway send back ACK for a write request."""
        if sid not in self._device_to_gw:
            _LOGGER.error("on_write_ack(): sid not found %s", sid)
            return
        self._device_to_gw[sid].on_write_ack(model, sid, data)

    def on_report(self, model, sid, data):
        """Called when a device sent a status report."""
        if sid not in self._device_to_gw:
            _LOGGER.warning("on_report(): sid not found %s", sid)
            return
        self._device_to_gw[sid].on_device_report(model, sid, data)

    def on_heartbeat(self, model, sid, data, gw_token):
        """Called when a heartbeat is received."""
        if sid not in self._device_to_gw:
            _LOGGER.warning("on_heartbeat(): sid not found %s", sid)
            return
        self._device_to_gw[sid].on_device_heartbeat(model, sid, data, gw_token)

    def subscribe(self, handle_new_gateway):
        """Subscribe to gateway events."""
        dispatcher.connect(handle_new_gateway, signal=AQARA_EVENT_NEW_GATEWAY, sender=self)

    def unsubscribe(self, handle_new_gateway):
        """Unsubscribe from gateway events."""
        dispatcher.disconnect(handle_new_gateway, signal=AQARA_EVENT_NEW_GATEWAY, sender=self)



"""Constants of the aqara protocol"""

MCAST_ADDR = "224.0.0.50"
MCAST_PORT = 4321

LISTEN_IP = "0.0.0.0"
LISTEN_PORT = 9898

GATEWAY_PORT = 9898

AQARA_DEVICE_HT = 'sensor_ht'
AQARA_DEVICE_MOTION = 'motion'
AQARA_DEVICE_MAGNET = 'magnet'
AQARA_DEVICE_SWITCH = 'switch'
AQARA_DEVICE_GATEWAY = 'gateway'

AQARA_SWITCH_ACTION_CLICK = 'click'
AQARA_SWITCH_ACTION_DOUBLE_CLICK = 'double_click'
AQARA_SWITCH_ACTION_LONG_CLICK_PRESS = 'long_click_press'
AQARA_SWITCH_ACTION_LONG_CLICK_RELEASE = 'long_click_release'

AQARA_MID_STOP = 10000

AQARA_ENCRYPT_IV = b'\x17\x99\x6d\x09\x3d\x28\xdd\xb3\xba\x69\x5a\x2e\x6f\x58\x56\x2e'

AQARA_DEFAULT_VOLTAGE = 9999 # set to a large number when unknown

AQARA_EVENT_NEW_GATEWAY = 'aqara_new_gateway'
AQARA_EVENT_NEW_DEVICE = 'aqara_new_device'
